from os import makedirs
from os.path import join, isdir
from warnings import warn_explicit
from typing import Any

from tiramisu import Config
from tiramisu.error import ValueWarning
from rougail import RougailConfig, RougailConvert
from .utils import MULTI_FUNCTIONS, CONFIGS, DOMAINS
from rougail.utils import normalize_family


ROUGAIL_NAMESPACE = 'general'
ROUGAIL_NAMESPACE_DESCRIPTION = 'Général'


async def set_linked_multi_variables(value: str,
                                     linked_server: str=None,
                                     variable_index: int=None,
                                     linked_returns: str=None,
                                     dynamic: str=None,
                                     **kwargs: dict,
                                     ) -> None:
    if value is not None and linked_server is not None and 'linked_value_0' not in kwargs:
        kwargs['linked_value_0'] = value
    elif not linked_server:
        linked_server = value
    if not linked_server:
        return
    if linked_server not in CONFIGS:
        warn_explicit(ValueWarning(f'cannot find linked server "{linked_server}"'),
                      ValueWarning,
                      __file__,
                      3,
                      )
        return
    config = CONFIGS[linked_server]['config']
    dico = {}
    variables = {}
    for key, kvalue in kwargs.items():
        try:
            index = int(key.rsplit('_', 1)[-1])
        except ValueError:
            raise Exception(f'unknown variable {key}')
        if kvalue is None and not kwargs.get(f'allow_none_{index}', False):
            return
        if key.startswith('linked_provider_'):
            path = await config.information.get('provider:' + kvalue, None)
            if not path:
                return
            if index not in variables:
                variables[index] = {'path': None, 'value': None, 'variable_index': False}
            variables[index]['path'] = path
        elif key.startswith('linked_value_'):
            index = int(key[13])
            if index not in variables:
                variables[index] = {'path': None, 'value': None, 'variable_index': False}
            variables[index]['value'] = kvalue
        elif key.startswith('variable_index_'):
            index = int(key[15])
            if index not in variables:
                variables[index] = {'path': None, 'value': None, 'variable_index': False}
            variables[index]['variable_index'] = True
        elif key.startswith('allow_none_'):
            pass
        else:
            raise AttributeError(f'unknown parameter {key}')
    await config.property.read_write()
    if not isinstance(variables[0]['value'], list):
        variables[0]['value'] = [variables[0]['value']]
    if dynamic:
        dynamic = normalize_family(dynamic)
    _dynamic = None
    try:
        if variables[0]['value']:
            for first_idx, first_value in enumerate(variables[0]['value']):
                slave_idxes = []
                if dynamic:
                    _dynamic = dynamic
                else:
                    _dynamic = normalize_family(first_value)
                for index in sorted(list(variables)):
                    path = variables[index]['path']
                    if '{suffix}' in path:
                        path = path.replace('{suffix}', _dynamic)
                    elif first_idx != 0:
                        continue
                    vvalue = variables[index]['value']
                    option = config.forcepermissive.option(path)
                    if not await option.option.isfollower():
                        #print('===>', path, vvalue, await option.option.ismulti(), await option.option.issubmulti())
                        multi = await option.option.ismulti()
                        if multi:
                            isleader = await option.option.isleader()
                            if not isinstance(vvalue, list):
                                vvalue = [vvalue]
    #                        elif isleader:
    #                            raise Exception('leader must not be a multi from now ...')
                            values = await option.value.get()
                            for val in vvalue:
                                if val not in values:
                                    if isleader:
                                        slave_idxes.append(len(values))
                                    values.append(val)
                                elif isleader:
                                    slave_idxes.append(values.index(val))
                            await option.value.set(values)
                            await option.owner.set('link')
                        else:
                            if isinstance(vvalue, list) and len(vvalue) == 1:
                                vvalue = vvalue[0]
                            await option.value.set(vvalue)
                            await option.owner.set('link')
                    else:
                        # print('===<', path, vvalue, await option.option.ismulti(), await option.option.issubmulti())
                        if not slave_idxes:
                            raise Exception('please declare the leader variable before the follower')
                        if variables[index]['variable_index']:
                            vvalue = vvalue[variable_index]
                        if not isinstance(vvalue, list):
                            vvalue = [vvalue] * len(slave_idxes)
    #                    if isinstance(vvalue, list) and not await option.option.issubmulti():
                        for idx, val in enumerate(vvalue):
                            option = config.forcepermissive.option(path, slave_idxes[idx])
                            await option.value.set(val)
                            await option.owner.set('link')
    except Exception as err:
        await config.property.read_only()
        raise err from err
    await config.property.read_only()
    if not dynamic:
        dynamic = _dynamic


    if linked_returns is not None:
        linked_variable = await config.information.get('provider:' + linked_returns, None)
        if not linked_variable:
            warn_explicit(ValueWarning(f'cannot find linked variable "{linked_returns}" in linked server "{linked_server}"'),
                          ValueWarning,
                          __file__,
                          0,
                          )
            return
        if dynamic:
            linked_variable = linked_variable.replace('{suffix}', normalize_family(dynamic))
        elif '{suffix}' in linked_variable:
            idx = CONFIGS[linked_server]['interface_index']
            linked_variable = linked_variable.replace('{suffix}', str(idx))
        ret = await config.forcepermissive.option(linked_variable).value.get()
    else:
        ret = get_ip_from_domain(linked_server)
    return ret


async def set_linked(linked_server: str,
                     linked_provider: str,
                     linked_value: str,
                     linked_returns: str=None,
                     dynamic: str=None,
                     ):
    if None in (linked_server, linked_provider, linked_value):
        return
    if linked_server not in CONFIGS:
        warn_explicit(ValueWarning(f'cannot find linked server "{linked_server}"'),
                      ValueWarning,
                      __file__,
                      0,
                      )
        return
    config = CONFIGS[linked_server]['config']
    path = await config.information.get('provider:' + linked_provider, None)
    if not path:
        warn_explicit(ValueWarning(f'cannot find provider "{linked_provider}" in linked server "{linked_server}"'),
                      ValueWarning,
                      __file__,
                      0,
                      )
        return
    await config.property.read_write()
    try:
        option = config.forcepermissive.option(path)
        if await option.option.ismulti():
            values = await option.value.get()
            if linked_value not in values:
                values.append(linked_value)
                await option.value.set(values)
                await option.owner.set('link')
        else:
            await option.value.set(linked_value)
            await option.owner.set('link')
    except Exception as err:
        await config.property.read_only()
        raise err from err
    await config.property.read_only()
    if linked_returns is not None:
        linked_variable = await config.information.get('provider:' + linked_returns, None)
        if not linked_variable:
            warn_explicit(ValueWarning(f'cannot find linked variable "{linked_returns}" in linked server "{linked_server}"'),
                          ValueWarning,
                          __file__,
                          0,
                          )
            return
    else:
        linked_variable = None
    if linked_variable is not None:
        if dynamic:
            linked_variable = linked_variable.replace('{suffix}', normalize_family(dynamic))
        elif '{suffix}' in linked_variable:
            idx = CONFIGS[linked_server]['interface_index']
            linked_variable = linked_variable.replace('{suffix}', str(idx))
        ret = await config.forcepermissive.option(linked_variable).value.get()
    else:
        ret = normalize_family(linked_value)
    return ret


async def get_linked_configuration(linked_server: str,
                                   linked_provider: str,
                                   dynamic: str=None,
                                   ):
    if linked_server not in CONFIGS:
        warn_explicit(ValueWarning(f'cannot find linked server "{linked_server}"'),
                      ValueWarning,
                      __file__,
                      1,
                      )
        return
    config = CONFIGS[linked_server]['config']
    path = await config.information.get('provider:' + linked_provider, None)
    if not path:
        warn_explicit(ValueWarning(f'cannot find variable "{path}" in linked server "{linked_server}"'),
                      ValueWarning,
                      __file__,
                      1,
                      )
        return
    if dynamic:
        path = path.replace('{suffix}', normalize_family(dynamic))
    try:
        return await config.forcepermissive.option(path).value.get()
    except AttributeError as err:
        warn_explicit(ValueWarning(f'cannot find get value of "{path}" in linked server "{linked_server}": {err}'),
                      ValueWarning,
                      __file__,
                      1,
                      )


class Empty:
    pass
empty = Empty()


async def set_linked_configuration(_linked_value: Any,
                                   linked_server: str,
                                   linked_provider: str,
                                   linked_value: Any=empty,
                                   dynamic: str=None,
                                   leader_provider: str=None,
                                   leader_value: Any=None,
                                   leader_index: int=None,
                                   ):
    if linked_value is not empty:
        _linked_value = linked_value
    linked_value = _linked_value
    if linked_server is None:
        return
    if linked_value is None or linked_server not in CONFIGS:
        warn_explicit(ValueWarning(f'cannot find linked server "{linked_server}"'),
                      ValueWarning,
                      __file__,
                      2,
                      )
        return
    config = CONFIGS[linked_server]['config']
    path = await config.information.get('provider:' + linked_provider, None)
    if not path:
        warn_explicit(ValueWarning(f'cannot find variable "{path}" in linked server "{linked_server}"'),
                      ValueWarning,
                      __file__,
                      2,
                      )
        return
    if dynamic:
        path = path.replace('{suffix}', normalize_family(dynamic))
    await config.property.read_write()
    try:
        if leader_provider is not None:
            leader_path = await config.information.get('provider:' + leader_provider, None)
            if not leader_path:
                await config.property.read_only()
                warn_explicit(ValueWarning(f'cannot find leader variable with leader_provider "{leader_provider}" in linked server "{linked_server}"'),
                              ValueWarning,
                              __file__,
                              2,
                              )
                return
            if dynamic:
                leader_path = leader_path.replace('{suffix}', normalize_family(dynamic))
            values = await config.forcepermissive.option(leader_path).value.get()
            if not isinstance(leader_value, list):
                leader_value = [leader_value]
            for lv in leader_value:
                if lv in values:
                    slave_idx = values.index(lv)
                    slave_option = config.forcepermissive.option(path, slave_idx)
                    if await slave_option.option.issubmulti():
                        slave_values = await slave_option.value.get()
                        if linked_value not in slave_values:
                            slave_values.append(linked_value)
                            await slave_option.value.set(slave_values)
                            await slave_option.owner.set('link')
                    else:
                        await slave_option.value.set(linked_value)
                        await slave_option.owner.set('link')
        else:
            option = config.forcepermissive.option(path, leader_index)
            if leader_index is None and await option.option.ismulti() and not isinstance(linked_value, list):
                values = await option.value.get()
                if linked_value not in values:
                    values.append(linked_value)
                    await option.value.set(values)
                    await option.owner.set('link')
            else:
                await option.value.set(linked_value)
                await option.owner.set('link')
    except AttributeError as err:
        if linked_provider == 'oauth2_external':
            raise ValueError(str(err)) from err
        pass
    except Exception as err:
        await config.property.read_only()
        raise err from err
    await config.property.read_only()


def get_ip_from_domain(domain):
    if not domain:
        return
    hostname, domainname = domain.split('.', 1)
    return DOMAINS[domainname][1][DOMAINS[domainname][0].index(hostname)]


async def load_machine_config(server_name: str,
                              datas: dict,
                              module_info: dict,
                              display_name,
                              ):
    optiondescription = {'set_linked': set_linked,
                         'set_linked_multi_variables': set_linked_multi_variables,
                         'get_linked_configuration': get_linked_configuration,
                         'set_linked_configuration': set_linked_configuration,
                         'get_ip_from_domain': get_ip_from_domain,
                         }
    cfg = RougailConfig.copy()
    module_info.servers.append(server_name)
    cfg['variable_namespace'] = ROUGAIL_NAMESPACE
    cfg['variable_namespace_description'] = ROUGAIL_NAMESPACE_DESCRIPTION
    if datas['module'] == 'host':
        cfg['tmpfile_dest_dir'] = datas['values'][f'{ROUGAIL_NAMESPACE}.host_install_dir'] + '/host/configurations/' + server_name
    cfg['templates_dir'] = module_info.templates_dir
    cfg['dictionaries_dir'] = module_info.dictionaries_dir
    cfg['functions_file'] = module_info.functions_file
    cfg['multi_functions'] = MULTI_FUNCTIONS
    cfg['extra_dictionaries'] = module_info.extra_dictionaries
    cfg['extra_annotators'].append('risotto.rougail')
    cfg['internal_functions'] = list(optiondescription.keys())
    try:
        eolobj = RougailConvert(cfg)
    except Exception as err:
        print(f'Try to load {module_info.modules}')
        raise err from err
    tiram_obj = eolobj.save(None)
#    if server_name == 'revprox.in.silique.fr':
#        print(tiram_obj)
    #cfg['patches_dir'] = join(test_dir, 'patches')
    try:
        exec(tiram_obj, None, optiondescription)
    except Exception as err:
        print(tiram_obj)
        raise Exception(f'unknown error when load tiramisu object {err}') from err
    config = await Config(optiondescription['option_0'], display_name=display_name)
    await config.property.read_write()
    try:
        add_srv = await config.option('machine.add_srv').value.get()
    except AttributeError:
        add_srv = False
    return {'config': config,
            'templates_informations': cfg,
            'interface_index': 0,
            'module_name': datas['module'],
            'add_srv': add_srv,
            }
